/*** 
	MACRO to implement MAX test of Zheng and Gastwirth 
	(Statistics in Medicine, Oct. 2006). 
	Developed by Dana Aeschliman <dana at statgen.org>
	for Marie-Pierre Dube's 
	Statistical Genetics Research Group at
	the Montreal Heart Institute 
	during Jan-March 2007

	In developing this MACRO I was often inspired by or just plain ripped off
	flow control ideas from Michael Friendly's biplot.sas
	Doc: http://www.math.yorku.ca/SCS/vcd/biplot.html
	The standing on shoulders of giants thing...
***/
%include "DO_QC.sas";
%macro MAXTEST(
dsn=, /* Input data set name */
aff_stat=case, /* Variable for affected status */
locus=locus, /* Variable giving locus name. Can be "locus". */
A1=a1, /* Variable giving name of first allele. Can be "a1". */
A2=a2, /* Variable giving name of second allele. Can be "a2". */
NUM_MVNS=10000, 
	/* NUM_MVNS: Note that we recommend that this be kept to at least 1e4 */
IDENT=id,
output_DSN=max_out, /*** output_DSN is only used when this MACRO (current MACRO) is
called from outside. ***/
use_lib=WORK, /*** use_lib is used for several things.
	It's the library that holds the raw data. It's the library
	where the quality control (QC) files &use_lib..bad_ccs, &use_lib..miss_or_diff_als, and
	&use_lib..snp_warnings will be deposited. It's the library
	where the results files max_data and eigen_vecs will be deposited in the 
	cases when use_lib^=WORK. ***/ 
skip_checks=F,
	/*** If skip_checks=T, then input data set must be exactly bi-allelic at 
	each locus and each SNP must have both cases and controls. As well, data
	set must have columns named "locus", "a1", "a2", "case", "ID". Basically, 
	skip_sets=T should mean that MAXTEST is being called from 
	SNP_step_down_p. Note that, when skip_checks=T, the MACRO vars aff_stat,
	locus, A1, A2, IDENT, and use_lib are not actually used.
	***/
keep_details=T,
	/*** keep_details=T means that the stats min_allele, Z_0, Z_05, Z_1, n0, n1, n2,
	corr_0_05, corr_05_1, corr_0_1 and  N will be calculated and ouput as well as 
	max_abs_zs and corresponding p-value.
	***/
/*********************************/
/*** graphic/ reports parameters ***/
title1=Results of Max Test of Freidlin et al. (2002),
	/*** User can put in more informative titles ***/
title2=MAX test results,
get_graphic=T, 
	/*** Flag telling if graphic should be produced. ***/
ps_graphic=40, ls_graphic=90,
	/*** Params controlling pagesize and linesize for graphic ***/
get_reports=T,
	/*** Flag telling if reports should be produced. ***/
ps_reports=35, ls_reports=90,
	/*** Params controlling pagesize and linesize for reports ***/
top_SNPs=25,
	/*** Print out a list of the top top_SNPs SNPs. Changing top_SNPs to a number
	less than 1 suppresses this output ***/
ps_top_SNPs=33, ls_top_SNPs=90
);

/*** Note that, every time MAXTEST is called with the parameter keep_details=T,
the following data sets get overwritten: &use_lib..bad_ccs, &use_lib..miss_or_diff_als, and
&use_lib..snp_warnings. Please use keep_details=T iff the current iteration treats 
the actual data.  ***/
%let dsn=%upcase(&dsn.); 
%let use_lib=%upcase(&use_lib.); 
%let skip_checks=%upcase(&skip_checks);
%let keep_details=%upcase(&keep_details);
%let top_SNPs=%sysfunc(floor(&top_SNPs));
%if &skip_checks=T %then %goto START;
%include "get_eigens.sas";
%include "get_pvals.sas";
%QC(dsn=&dsn, use_lib=&use_lib, FROM_MACRO=MAX_TEST, out_dsn=good_data);
%if &_abort_=1 %then %goto done;
%start:
/*** Calculate the 3 U stats, their variances and the Z stats. Calculate the
v's and corrs (from the APPENDIX A of Zheng and Gastwirth). Calculate the
MAX ***/
data tmp_max_data(drop=a1 a2 case ID);
/*** If there is no other MACRO calling this one... ***/
%if &keep_details=T and &skip_checks=F %then %do;
set good_data; 
by locus_no;
%end;
/*** If there is another MACRO calling this one and it's the first time through... ***/
%if &keep_details=T and &skip_checks=T %then %do;
set &dsn; 
BY locus_no;
%end;
/*** If there is another MACRO calling this one and it's NOT the first time through... ***/
/*** Note that, because of the way that we resample (as of Mar 30, 2007 we are permuting
in STEPDOWN and bootstrapping isn't an option) and the way that we
treat missing values, the numbers R and S
and r0, r1, r2, s0, s1, and s2 may change but N, n0, n1, n2, min_allele, maj_allele
can't change. Which group (R or S) gets assigned the missing value may change
but we are always dealing with the same set of genotypes. ***/
%if &keep_details=F and &skip_checks=T %then %do;
merge &dsn(in=A) &use_lib..max_data(keep=locus_no min_allele maj_allele n0 n1 n2 n in=B);
BY locus_no;
if A and B;
%end;
retain r0 r1 r2 s0 s1 s2; 
if first.locus_no then do;
	r0=0; r1=0; r2=0; s0=0; s1=0; s2=0;
end;
/*** Get set of 2x3 tables for good_data ***/
/*** I think that this is at least as
fast as using proc_freq and ODS tables. Might think more about this
later ***/
/*** If it's the first time through, we will have arbitrarily assigned 
min_allele in the QC scripts accessed above. We make sure below that we count 
alleles with smaller freq. Otherwise, when it's not the first time through, 
we've loaded min_allele above ***/
if case=1 then do;
	if a1^=min_allele and a2^=min_allele then r0=r0+1;
	else if (a1=min_allele and a2^=min_allele) or (a1^=min_allele and a2=min_allele) 
		then r1=r1+1;
	else if a1=min_allele and a2=min_allele then r2=r2+1;
end;
else do;
	if a1^=min_allele and a2^=min_allele then s0=s0+1;
	else if (a1=min_allele and a2^=min_allele) or (a1^=min_allele and a2=min_allele) 
		then s1=s1+1;
	else if a1=min_allele and a2=min_allele then s2=s2+1;
end;

if last.locus_no then do;
/*** Small sample correction. This is done every time. We could test if there are any
0's in our CC stat x genotype table first but that would take a bit of time. As well,
in cases where the no. of missings is >= R or S we could have a situation where, due to 
resampling and subsequent checking that some 0's exist and add'ns of 1/2, the marginals 
change. So, although we can anticipate some small loss in power, we add the 1/2's. ***/
	r0=r0+0.5; r1=r1+0.5; r2=r2+0.5;
	s0=s0+0.5; s1=s1+0.5; s2=s2+0.5;
%if &keep_details=T %then %do;
/*** &keep_details=T iff the current iteration treats the actual data. ***/
n0=r0+s0; n1=r1+s1; n2=r2+s2; N=n0+n1+n2;
%end;
output;
end; /*** end of "if last.locus_no then do;" ***/
run;

/* I'm guessing that it's less time consuming to keep even things that are easy to
	calculate, like n's, N, etc. than to recalculate them every time. */
%if &keep_details=T %then %do;
data tmp_max_data(drop=U_rec U_add U_dom tmp tmp2 sqrt_scale_fac);
set tmp_max_data;
/* If it's the first time through, make sure that we're counting minor allele */
	frac_min=(n1+2*n2)/(2*N);
	if frac_min>0.5 then do;
		tmp=min_allele; min_allele=maj_allele; maj_allele=tmp;
/*** I think that it's necessary to use another var here, tmp2, because 
		tmp is already defined as an integer (???) ***/
		tmp2=r0; r0=r2; r2=tmp2;
		tmp2=s0; s0=s2; s2=tmp2;
		tmp2=n0; n0=n2; n2=tmp2;
		frac_min=1-frac_min;
	end;
/*** Feb. 12. We are currently programming MAX test, not MAX_CC ***/
	p0=n0/n; p1=n1/n; p2=n2/n; 
	sqrt_v0=(p2*(1-p2))**0.5; 
	sqrt_v1=(p0*p1+p1*p2+4*p0*p2)**0.5;
	sqrt_v2=(p0*(1-p0))**0.5;
	corr_0_05=p2*(2*p0+p1)/(sqrt_v0*sqrt_v1);
	corr_05_1=p0*(2*p2+p1)/(sqrt_v2*sqrt_v1);
	corr_0_1=p0*p2/(sqrt_v0*sqrt_v2);
	R=r0+r1+r2;
	S=N-R;
	phi=R/N; 
/*** Appendix A of Zheng and Gastwirth ***/
	sqrt_scale_fac=((1-phi)*phi*N)**0.5;
	sqrt_var_U_rec=sqrt_scale_fac*sqrt_v0;
	/*** Note that sqrt(0.25)=0.5 ***/
	sqrt_var_U_add=0.5*sqrt_scale_fac*sqrt_v1;
	sqrt_var_U_dom=sqrt_scale_fac*sqrt_v2;
%end; /*** end of "if keep_details=T then do" ***/
%else %do; 
	%let _some_R_changed_=0; 
		/*** This is flag for telling, in presence of missing values,
		if we have some Rs and Ss that have changed. So, _some_R_changed_=1 
		implies that we have some missing values and that
		R and S were changed due to permuting with missing values. Thus,
		we will need to recompute some eigen info. 
		***/
	%if &_hay_missing_=0 %then %do;
data tmp_max_data(drop=U_rec U_add U_dom);
merge tmp_max_data &use_lib..max_data(keep=phi r s n0 n1 n2 N
	sqrt_var_U_rec sqrt_var_U_add sqrt_var_U_dom corr_0_05  corr_0_1 corr_05_1
	locus_no) end=last;
	by locus_no;
	%end;
	%else %do; 
		/*** This is flag for telling, in presence of missing values,
		if we have some Rs and Ss that have changed. ***/
data tmp_max_data(drop=U_rec U_add U_dom sqrt_scale_fac some_R_changed tmp );
merge tmp_max_data &use_lib..max_data(keep=phi r s n0 n1 n2 N
	sqrt_var_U_rec sqrt_var_U_add sqrt_var_U_dom corr_0_05  corr_0_1 corr_05_1
	sqrt_v0 sqrt_v1 sqrt_v2 locus_no ) end=last;
	by locus_no;
/*******************************************************************************/
/*** Calculate some params that we only need the 1st time through or when the resampling
in presence of missing values has caused R and S to change from what we saw with the
observed data. ***/
/*******************************************************************************/
retain some_R_changed 0;
/*** In cases where it's not the 1st time through, there's some doubt as to 
whether we need to do this or not. ***/
recompute_params=0;
/*** recompute_params tells if, for this particular SNP, we recompuet the params ***/
if this_locus=1 then do; /*** This locus has
at least one missing value. Note that the fact that the locus has a missing
value doesn't necessarily imply that R and S have changed due to the permutation;
for example, the missing value on the observed data may belong to an R and may be 
reassigned to an R. ***/
tmp=r0+r1+r2;
	if R^=tmp then do;
		R=tmp;
		S=N-R;
		phi=R/N; 
/*** Appendix A of Zheng and Gastwirth ***/
		sqrt_scale_fac=((1-phi)*phi*N)**0.5;
		sqrt_var_U_rec=sqrt_scale_fac*sqrt_v0;
	/*** Note that sqrt(0.25)=0.5 ***/
		sqrt_var_U_add=0.5*sqrt_scale_fac*sqrt_v1;
		sqrt_var_U_dom=sqrt_scale_fac*sqrt_v2;
	end; /*** end of if R^=tmp then do ***/
end; /*** end of if this_locus=1; ***/ 
	%end; /*** End of %if &_hay_missing_=0 %then %do;...end; else do ... ***/
%end; /*** end of "if keep_details=T then do...else do ... " ***/
	U_rec=(1-phi)*r2-phi*s2; 
	U_add=(1-phi)*(0.5*r1+r2)-phi*(0.5*s1+s2); 
	U_dom=(1-phi)*(r-r0)-phi*(s-s0); 

	Z_0=U_rec/sqrt_var_U_rec;
	Z_05=U_add/sqrt_var_U_add;
	Z_1=U_dom/sqrt_var_U_dom;
	max_abs_zs=max(abs(Z_0), abs(Z_05), abs(Z_1));
	run;

/******************************************************************************************/
/*** 
*** Calculate eigenvalues, calculate eigenvectors, count number of distinct SNPs ***
***/
/******************************************************************************************/
%if &keep_details=T %then %do;
	%if &skip_checks=F %then %do;
/*** We only need to prepare eigen vals and vecs if we're not going to implement Alg 4.1 ***/
		%get_eigens(dsn=tmp_max_data);
	%end;
/*** If appropriate, deposit eigen_vecs and max_data onto &use_lib. ***/
	%if %UPCASE(&use_lib.)^=WORK %then %do;
		%if &skip_checks=F %then %do;
			data &use_lib..eigen_vecs; 
			set eigen_vecs(drop=corr_0_05  corr_0_1 corr_05_1); run;
		%end;
data &use_lib..max_data; 
	set tmp_max_data; 
run;
	%END;
	%else %do;
data max_data; 
	set tmp_max_data; 
run;
	%end;
%end; /*** end of %if &keep_details=T %then %do; ***/
/******************************************************************************************/
/*** Note that if &skip_checks=T then we don't write to hard disk ***/
%if &skip_checks=T %then %do;
		data &output_DSN; 
		set tmp_max_data(keep=locus_no max_abs_zs); 
		run;
%end;
/**************************************************************************************/
/*** Clean up data sets prior to exiting or to generating random norms and then 
	crunching one record at a time.
/**************************************************************************************/

proc datasets; 
delete tmp_max_data
/*** Also delete set "good_data" if MAX isn't being called from outside. ***/
%if &skip_checks=F %then %do;
	good_data
%end; 
;
run;
quit;
/*** If we're calling current MACRO from outside, we're done here. ***/
%if &skip_checks=T %then %goto done;
/***********************************************************************/

/*** Generate matrix of standard, uncorrelated random normals ***/

/***********************************************************************/
data norms(keep=n1 n2 n3); 
/*** We don't need to have separate streams so the args for n2 and n3 are just 
dummy vals ***/
do i=1 to &NUM_MVNS.;
	n1=rannor(0);
	n2=rannor(1);
	n3=rannor(1);
	output;
end;
run;

/**************************************************************************************/
/*** Start crunching one record at a time. ***/
/**************************************************************************************/
%do i=1 %to &SNP_count;
/***** Peel off one SNP. Set up the columns of sqrt of the covariance matrix
for this SNP. *****/

	data _NULL_; 
		set eigen_vecs(keep=locus_no b1_1 b2_1 b3_1 b1_2 b2_2 b3_2 
		b1_3 b2_3 b3_3 eigen_val1 eigen_val2 eigen_val3 max_abs_zs
		firstobs=&i obs=&i);
	call symput('_locus_no_', locus_no);
	call symput('_max_abs_zs_', max_abs_zs);
	call symput('_eigen_val3_', eigen_val3);
	call symput('f11', b1_1*eigen_val1**0.5); 
	call symput('f21', b2_1*eigen_val1**0.5); 
	call symput('f31', b3_1*eigen_val1**0.5); 
 
	call symput('f12', b1_2*eigen_val2**0.5); 
	call symput('f22', b2_2*eigen_val2**0.5); 
	call symput('f32', b3_2*eigen_val2**0.5); 

	if eigen_val3^=0 then do;
		call symput('f13', b1_3*eigen_val3**0.5); 
		call symput('f23', b2_3*eigen_val3**0.5); 
		call symput('f33', b3_3*eigen_val3**0.5); 
	end;
	run;

%global _p_val_ _num_p_ ;
%let _num_p_=0;
%let NUM_MVNs_used=200;
%if %sysevalf(&NUM_MVNs<200) %then %let NUM_MVNs_used=&NUM_MVNs;
%pvals(STATEMENT1= obs=&NUM_MVNs_used, statement2= &NUM_MVNs_used);
%if %sysevalf(&_p_val_<=0.2) and &NUM_MVNs>200 %then %do;
	%let NUM_MVNs_used=1000;
	%if %sysevalf(&NUM_MVNs<1000) %then %let NUM_MVNs_used=&NUM_MVNs;
	%pvals(STATEMENT1=firstobs=201 obs=&NUM_MVNs_used, statement2=&NUM_MVNs_used);
%end;

%if %sysevalf(&_p_val_<=0.1) and &NUM_MVNs>1000 %then %do;
	%let NUM_MVNs_used=5000;
	%if %sysevalf(&NUM_MVNs<5000) %then %let NUM_MVNs_used=&NUM_MVNs;
	%pvals(STATEMENT1=firstobs=1001 obs=&NUM_MVNs_used, statement2=&NUM_MVNs_used);
%end;

%if %sysevalf(&_p_val_<=0.05) and &NUM_MVNs>5000 %then %do;
	%pvals(STATEMENT1=firstobs=5001, statement2=&NUM_MVNs);
%end;

	%if &i=1 %then %do;
		data keep_ps; set pval_for_single; run;
	%end;
	%else %do;
		proc append base=keep_ps data=pval_for_single; run; quit;
	%end;
%end; /*** end: %do i=1 %to &SNP_count; ***/

	data tmp;
	merge &use_lib..max_data keep_ps(keep=p_val locus_no) ;
	by locus_no;
	run;
/*** Add back in SNP names. ***/
	data &use_lib..max_data;
	merge tmp  &use_lib..locus_key;
	by locus_no;
	run;
/*** Final clean up of data sets. ***/
proc datasets; delete keep_ps norms pval_for_single tmp
%if %UPCASE(&use_lib.)^=WORK %then %do;
EIGEN_VECS
%end;
; run; quit;
/*******************************************************/
title1 "&title1";
title2 "&title2";
%let SNP_COUNT=%trim(%left(&SNP_COUNT));
/**** Do final reporting ***/
/*** First, do 2 reports: Report 1 shows the CC stat vs genotype table for each
locus. Report 2 shows the results of the MAX test. ***/
%if &get_reports=T %then %do;
title3 "A total of &SNP_count SNPs were analyzed.";
title4 'Numbers used in CC Stat by Genotype Table';
title5 "(Prior to add'n of 1/2 to each cell)";
options ps=&ps_reports ls=&ls_reports;

data tmp(keep=locus c0 c1 c2 rowtot rowname MAF); 
set &use_lib..max_data(keep=locus r0 r1 r2 s0 s1 s2 n1 n2 n);
/*** Note that we formerly added 1/2 to each of these values ***/
	MAF=(n1+2*n2-1.5)/(2*(N-3));
	c0=r0-0.5; c1=r1-0.5; c2=r2-0.5;
	rowtot=c0+c1+c2;
	rowname='Case';
	output;
	c0=s0-0.5; c1=s1-0.5; c2=s2-0.5;
	rowtot=c0+c1+c2;
	rowname='Control';
	output;
	run;

proc report data=tmp nofs split="*"; 
column locus MAF ('-Genotype Distribution Table-*Minor Allele Count' 
	rowname c0 c1 c2 rowtot);
define locus / group;
define MAF / group format=3.2 'MAF*---';
define rowname / display width=7 'CC stat*---';
define c0 / analysis sum format=7. 'Count=0*---';
define c1 / analysis sum  format=7. 'Count=1*---';
define c2 / analysis sum  format=7. 'Count=2*---';
define rowtot / analysis sum format=7.  'Row Tot*---';
break after locus / ol summarize skip suppress;
run; quit;


title3 "A total of &num_mvns Monte Carlo samples were drawn for each SNP.";
title4 "A total of &SNP_count SNPs were analyzed.";
proc report data=&use_lib..max_data(keep=locus z_0 z_05 z_1 max_abs_zs p_val) nofs split="*"; 
column locus ('-Details of MAX test-' 
('-Components-' z_0 z_05 z_1) max_abs_zs p_val);
define locus / display ;
define z_0 / display format=7.2 'Rec';
define z_05 / display format=7.2 'Add';
define z_1 / display format=7.2 'Dom';
define max_abs_zs / display format=7.2 'MAX';
define p_val / display format=11.5 'p-val(MAX)';
run; quit;
%end; /*** end of "%if &get_reports=T ***/
/*********************************************************************/
/*** Now do 2 graphs. ***/
/*** Try to give an idea of population substructure or of true discovery rate ***/
/*** Make Probability-Probability Plot of observed p-values vs U(0,1) ***/
%if &get_graphic=T %then %do;
options ps=&ps_graphic. ls=&ls_graphic.;

data tmp(keep=log_runi); 
log_min_p=0; den=&SNP_count+1;
do i=1 to &SNP_count;
log_runi=log10(i/den);
log_min_p=min(log_min_p, log_runi);
output;
end;
call symput('_log_min_p_', log_min_p);
run;
proc sort; by log_runi; run;
data tmp2(keep=log_p_val); set &use_lib..max_data(keep=p_val) end=last; 
retain log_min_p &_log_min_p_.;
log_p_val=log10(p_val);
log_min_p=min(log_min_p, log_p_val);
if last then do;
	call symput('_log_min_p_', log_min_p);
	log_int_min=floor(log_min_p);
	tf_log_int_min=(3/4)*log_int_min;
	oh_log_int_min=(1/2)*log_int_min;
	of_log_int_min=(1/4)*log_int_min;
	call symput('_log_int_min_', log_int_min);
	call symput('_tf_log_int_min_', tf_log_int_min);
	call symput('_oh_log_int_min_', oh_log_int_min);
	call symput('_of_log_int_min_', of_log_int_min);
end;
run;
data tmp2b; set tmp2(where=(log_p_val=.));
log_p_val=input(trim(left(symget('_log_int_min_'))), best8.)+0.1;
run;
data tmp2; set tmp2(where=(log_p_val^=.)) tmp2b; run;
proc sort ; by log_p_val; run;

data tmp3(keep=xy_p); 
log_min_p=input(trim(left(symget('_log_min_p_'))), best8.);
SNP_count=input(trim(left(symget('SNP_count'))), best8.);
do i=1 to SNP_count;
	xy_p=log_min_p+(i-1)*(0-log_min_p)/SNP_count;
	output;
end;
run;

data tmp; merge tmp tmp2; run;
data tmp; merge tmp tmp3; run;
ods pdf file="g1.pdf";
proc plot data=tmp nolegend;
   plot log_p_val*log_runi='*' 
        xy_p*xy_p='o' / overlay box 
			haxis=&_log_int_min_. &_tf_log_int_min_. &_oh_log_int_min_. &_of_log_int_min_. 0
			vaxis=&_log_int_min_. &_tf_log_int_min_. &_oh_log_int_min_. &_of_log_int_min_. 0;
	title3 "A total of &SNP_count SNPs were analyzed.";
	title4 'Probability-Probability Plot';
	title5 'log10 of p-values from MAX test vs log10 of quantiles of U(0,1) dist';
	
run;
quit;
ods pdf close;
%end; /** end of "%if get_graphic_=T %then %do ***/
/*********************************************************************/
/*** Now print out top SNPs ***/
%if &top_SNPs>=1 %then %do;
options ps=&ps_top_SNPs ls=&ls_top_SNPs;
proc sort data=&use_lib..max_data(keep=locus p_val z_0 z_05 z_1 max_abs_zs) out=tmp; 
by p_val; run;

title3 "A total of &num_mvns Monte Carlo samples were drawn for each SNP.";
title4 "A total of &SNP_count SNPs were analyzed.";
title5 "List of &top_SNPs. SNPs most likely to be associated with the phenotype.";
proc report data=tmp(obs=&top_SNPs.) nofs split="*"; 
column locus ('-Details of MAX test-' 
('-Components-' z_0 z_05 z_1) max_abs_zs p_val);
define locus / display ;
define z_0 / display format=7.2 'Rec';
define z_05 / display format=7.2 'Add';
define z_1 / display format=7.2 'Dom';
define max_abs_zs / display format=7.2 'MAX';
define p_val / display format=11.5 'p-val(MAX)';
run; quit;
%end; /*** end of %if &top_SNPs>=1 %then %do; ***/
/*** Clean-up ***/
proc datasets; delete tmp tmp2 tmp3; run; quit;
/*** End of section outputting final results for MAX test ***/
/*************************************************************/
%done:
%MEND; /* END OF MACRO MAXTEST */





